# Generated by Django 3.1.7 on 2021-08-06 08:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0155_merge_20210728_1430'),
    ]

    operations = [
        migrations.AlterField(
            model_name='opportunity',
            name='name',
            field=models.CharField(db_index=True, max_length=100),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='summary',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
    ]
