# Generated by Django 4.0.4 on 2022-08-08 07:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0166_rename_initial_ticket_initial_description_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='is_issue_flag',
            field=models.BooleanField(default=False),
        ),
    ]
