# Generated by Django 4.0.7 on 2023-03-08 17:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0168_ticket_contact_email_address_ticket_contact_name_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='connectwiseboard',
            name='time_entry_discussion_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='connectwiseboard',
            name='time_entry_internal_analysis_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='connectwiseboard',
            name='time_entry_resolution_flag',
            field=models.BooleanField(default=False),
        ),
    ]
