# Generated by Django 4.0.7 on 2022-10-07 10:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0167_alter_ticket_is_issue_flag'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='contact_email_address',
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='contact_name',
            field=models.CharField(blank=True, max_length=62, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='contact_phone_extension',
            field=models.CharField(blank=True, max_length=15, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='contact_phone_number',
            field=models.CharField(blank=True, max_length=20, null=True),
        ),
    ]
