# Generated by Django 4.0.10 on 2023-06-28 12:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0175_rename_accountmanagerflag_companyteam_account_manager_flag_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='billing_method',
            field=models.CharField(blank=True, choices=[('ActualRates', 'Actual Rates'), ('FixedFee', 'Fixed Fee'), ('NotToExceed', 'Not To Exceed'), ('OverrideRate', 'Override Rate')], max_length=50, null=True),
        ),
    ]
