# Generated by Django 2.1 on 2019-07-05 10:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0086_merge_20190507_1402'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='automatic_email_cc_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='automatic_email_contact_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='automatic_email_resource_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='bill_time',
            field=models.CharField(blank=True, choices=[('Billable', 'Billable'), ('DoNotBill', 'Do Not Bill'), ('NoCharge', 'No Charge'), ('NoDefault', 'No Default')], max_length=20, null=True),
        ),
    ]
