# Generated by Django 2.0 on 2018-08-14 10:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0065_auto_20180809_1124'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sla',
            name='based_on',
            field=models.CharField(choices=[('MyCalendar', 'My Company Calendar'), ('Customer', "Customer's Calendar"), ('AllHours', '24 Hours'), ('Custom', 'Custom Calendar')], default='MyCalendar', max_length=50),
        ),
    ]
