# Generated by Django 2.1.14 on 2020-06-24 10:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0118_auto_20200623_1016'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='scheduled_end',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='project',
            name='scheduled_start',
            field=models.DateField(blank=True, null=True),
        ),
    ]
