# Generated by Django 4.0.4 on 2022-06-22 16:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0163_ticket_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='source_tickets', to='djconnectwise.source'),
        ),
    ]
