# Generated by Django 3.1.2 on 2020-10-16 08:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0134_auto_20201015_1804'),
    ]

    operations = [
        migrations.AddField(
            model_name='activity',
            name='udf',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='opportunity',
            name='udf',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='project',
            name='udf',
            field=models.JSONField(blank=True, null=True),
        ),
    ]
