# Generated by Django 3.1.5 on 2021-06-10 17:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0149_auto_20210427_2236'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='estimated_start_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='wbs_code',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]
