# Generated by Django 4.0.4 on 2022-07-29 12:55

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0164_alter_ticket_source'),
    ]

    operations = [
        migrations.CreateModel(
            name='SourceTracker',
            fields=[
            ],
            options={
                'db_table': 'djconnectwise_source',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djconnectwise.source',),
        ),
    ]
