# Generated by Django 4.0.10 on 2023-06-20 17:58

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0174_companyteamrole_companyteam_companyteamroletracker_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='companyteam',
            old_name='accountManagerFlag',
            new_name='account_manager_flag',
        ),
        migrations.RenameField(
            model_name='companyteam',
            old_name='salesFlag',
            new_name='sales_flag',
        ),
        migrations.RenameField(
            model_name='companyteam',
            old_name='techFlag',
            new_name='tech_flag',
        ),
        migrations.RenameField(
            model_name='companyteamrole',
            old_name='accountManagerFlag',
            new_name='account_manager_flag',
        ),
        migrations.RenameField(
            model_name='companyteamrole',
            old_name='salesFlag',
            new_name='sales_flag',
        ),
        migrations.RenameField(
            model_name='companyteamrole',
            old_name='techFlag',
            new_name='tech_flag',
        ),
    ]
