# Generated by Django 4.0.10 on 2023-05-19 19:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0170_ticket_predecessor_closed_flag'),
    ]

    operations = [
        migrations.CreateModel(
            name='CompanyNoteType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('identifier', models.CharField(max_length=50)),
                ('defaultFlag', models.BooleanField()),
            ],
            options={
                'ordering': ('name',),
            },
        ),
    ]
